from .download import download_zip, download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_zip(xz_url + "assets.zip",down_path)


import bpy
import os
from bpy.types import Panel, Operator
from .translation.translate import GetTranslationDict


bl_info = {
    "name": "Handy Curve Profile",
    "author": "Young",
    "version": (1, 4, 5),
    "blender": (3, 5, 0),
    "location": "View3D > N-Panel",
    "description": "",
    "warning": "",
    "doc_url": "",
}

def set_squash(self, value):
    self["curve_profile_squash"] = value

def set_stretch(self, value):
    self["curve_profile_stretch"] = value

def set_curve_profile_scale(self, value):
    self["curve_profile_scale"] = value

def set_profiles_updated(self, value):
    self["profiles_updated"] = value

def Update_curve_profile_squash(self, context):
    curve_profile_squash = context.scene.handy_curve_profile_props.curve_profile_squash
    objs = bpy.context.selected_objects

    for ob in objs:
        if ob.type == "MESH":
            if ob.get('HCP_Pro'): 
                ob.scale[1] = curve_profile_squash
                ob.scale[2] = curve_profile_squash
            elif ob.get('HCP_End') and ob.get('HCP_Type'):
                if ob['HCP_Type']==1:
                    ob.scale[1] = curve_profile_squash
                    ob.scale[2] = curve_profile_squash
    return

def Update_curve_profile_stretch(self, context):
    curve_profile_stretch = context.scene.handy_curve_profile_props.curve_profile_stretch
    objs = bpy.context.selected_objects
    for ob in objs:
        if ob.type == "MESH":
            if ob.get('HCP_Pro'):
                ob.scale[0] = curve_profile_stretch
            elif ob.get('HCP_End') and ob.get('HCP_Type'):
                if ob['HCP_Type']==1:
                    ob.scale[0] = curve_profile_stretch

    return

def Update_curve_profile_scale(self, context):
    handy_curve_profile_props = context.scene.handy_curve_profile_props
    curve_profile_scale = handy_curve_profile_props.curve_profile_scale
    objs = bpy.context.selected_objects

    for ob in objs:
        if ob.type == "MESH":
            if ob.get('HCP_Pro') or ob.get('HCP_End'):
                ob.scale[0] = curve_profile_scale
                ob.scale[1] = curve_profile_scale
                ob.scale[2] = curve_profile_scale

    set_stretch(handy_curve_profile_props, curve_profile_scale)
    set_squash(handy_curve_profile_props, curve_profile_scale)

    return


def Update_curve_profile_offset(self, context):
    curve_profile_offset = context.scene.handy_curve_profile_props.curve_profile_offset
    objs = bpy.context.selected_objects
    for ob in objs:
        if ob.type == "MESH":
            if ob.get('HCP_Pro') or ob.get('HCP_End'):
                if ob.modifiers.get('HCP_Curve'):
                    cur_loc_x = ob.modifiers['HCP_Curve'].object.location[0]
                    ob.location[0] = curve_profile_offset +cur_loc_x

    return

def Update_curve_profile_rotation(self, context):
    curve_profile_rotation = context.scene.handy_curve_profile_props.curve_profile_rotation
    curve_profile_rotation_converted = curve_profile_rotation * 3.1415/180
    objs = bpy.context.selected_objects
    for ob in objs:
        if ob.type == "MESH":
            if ob.get('HCP_Pro') or ob.get('HCP_End'):
                ob.rotation_euler[0] = curve_profile_rotation_converted

    return

def set_array_type_fit(self, value):
    self["array_type_fit"] = value

def set_array_type_manual(self, value):
    self["array_type_manual"] = value

def Update_array_type_fit(self, context):
    handy_curve_profile_props = context.scene.handy_curve_profile_props
    array_type_fit = handy_curve_profile_props.array_type_fit
    objs = bpy.context.selected_objects

    if array_type_fit:
        set_curve_profile_scale(handy_curve_profile_props, 1)
        set_squash(handy_curve_profile_props, 1)
        set_stretch(handy_curve_profile_props, 1)
        set_array_type_manual(handy_curve_profile_props, False)
        for obj in objs:
            if obj.type == "MESH":
                if obj.modifiers.get('HCP_Array'):
                    obj.modifiers['HCP_Array'].fit_type = 'FIT_CURVE'
        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)

    else:
        set_array_type_manual(handy_curve_profile_props, True)
        for obj in objs:
            if obj.type == "MESH":
                if obj.modifiers.get('HCP_Array'):
                    obj.modifiers['HCP_Array'].fit_type = 'FIXED_COUNT'

    return


def Update_array_type_manual(self, context):
    handy_curve_profile_props = context.scene.handy_curve_profile_props
    array_type_manual = handy_curve_profile_props.array_type_manual
    objs = bpy.context.selected_objects

    if array_type_manual:
        set_array_type_fit(handy_curve_profile_props, False)
        for obj in objs:
            if obj.type == "MESH":
                if obj.modifiers.get('HCP_Array'):
                    obj.modifiers['HCP_Array'].fit_type = 'FIXED_COUNT'
    else:
        set_curve_profile_scale(handy_curve_profile_props, 1)
        set_squash(handy_curve_profile_props, 1)
        set_stretch(handy_curve_profile_props, 1)
        set_array_type_fit(handy_curve_profile_props, True)
        for obj in objs:
            if obj.type == "MESH":
                if obj.modifiers.get('HCP_Array'):
                    obj.modifiers['HCP_Array'].fit_type = 'FIT_CURVE'
        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)

    return


def Update_segment_distance(self, context):
    segment_distance = context.scene.handy_curve_profile_props.segment_distance
    objs = bpy.context.selected_objects

    for obj in objs:
        if obj.modifiers.get('HCP_Array'):
            obj.modifiers['HCP_Array'].relative_offset_displace[0] = segment_distance

    return


def Update_segment_count(self, context):
    segment_count = context.scene.handy_curve_profile_props.segment_count
    objs = bpy.context.selected_objects

    for obj in objs:
        if obj.modifiers.get('HCP_Array'):
            obj.modifiers['HCP_Array'].count = segment_count
    return



def blend_files_non_end_callback(self, context):
    items = []

    pref_path = bpy.context.preferences.addons[__package__].preferences.filepath
    directory = ''
    if pref_path == '':
        script_file = os.path.realpath(__file__)
        directory = os.path.join(os.path.dirname(script_file),"assets")
    else:
        directory = pref_path


    files = os.listdir(directory)
    blend_files = [file for file in files if file.endswith(".blend") and not "_end" in os.path.splitext(file)[0]]
    for file in blend_files:
        
        items.append((file, file, file))

    return items


def blend_files_end_callback(self, context):
    items = []

    pref_path = bpy.context.preferences.addons[__package__].preferences.filepath
    directory = ''
    if pref_path == '':
        script_file = os.path.realpath(__file__)
        directory = os.path.join(os.path.dirname(script_file),"assets")
    else:
        directory = pref_path


    files = os.listdir(directory)
    blend_files = [file for file in files if file.endswith(".blend") and "_end" in os.path.splitext(file)[0]]
    for file in blend_files:
        
        items.append((file, file, file))

    return items


def all_blend_files_callback(self, context):
    items = []

    pref_path = bpy.context.preferences.addons[__package__].preferences.filepath
    directory = ''
    if pref_path == '':
        script_file = os.path.realpath(__file__)
        directory = os.path.join(os.path.dirname(script_file),"assets")
    else:
        directory = pref_path


    files = os.listdir(directory)
    blend_files = [file for file in files if file.endswith(".blend")]
    for file in blend_files:
        
        items.append((file, file, file))

    return items



class HandyCurveProfileProperties(bpy.types.PropertyGroup):
    segment_count:  bpy.props.IntProperty(
        name='segment_count', default=3, min=1, soft_max=20, update=Update_segment_count)

    segment_distance:  bpy.props.FloatProperty(
        name='segment_distance', default=1, min=1, soft_max=10, update=Update_segment_distance)

    bone_cyclic_toggle: bpy.props.BoolProperty(
        name='bone_cyclic_toggle', default=False)

    profiles_updated: bpy.props.BoolProperty(
        name='profiles_updated', default=False)

    radius_tilt: bpy.props.BoolProperty(
        name='radius_tilt', default=False)

    misc_toggle: bpy.props.BoolProperty(
        name='misc_toggle', default=False)

    array_type_fit: bpy.props.BoolProperty(
        name='array_type_fit', default=True, update=Update_array_type_fit)

    array_type_manual: bpy.props.BoolProperty(
        name='array_type_manual', default=False, update=Update_array_type_manual)

    curve_profile_rotation: bpy.props.FloatProperty(
        name='curve_profile_offset', default=0, soft_min=-180, soft_max=180, update=Update_curve_profile_rotation)

    curve_profile_offset: bpy.props.FloatProperty(
        name='curve_profile_offset', default=0, soft_min=-5, soft_max=5, update=Update_curve_profile_offset)

    curve_profile_squash: bpy.props.FloatProperty(
        name='curve_profile_squash', default=1, min=0.01, soft_max=5, update=Update_curve_profile_squash)
    curve_profile_stretch: bpy.props.FloatProperty(
        name='curve_profile_stretch', default=1, min=0.01, soft_max=5, update=Update_curve_profile_stretch)
    curve_profile_scale: bpy.props.FloatProperty(
        name='curve_profile_scale', default=1, min=0.01, soft_max=5, update=Update_curve_profile_scale)


    profile_mesh_blend_files: bpy.props.EnumProperty(
        name="",
        description="Blender files from folder",
        items= blend_files_non_end_callback
        )


    end_mesh_blend_files: bpy.props.EnumProperty(
        name="",
        description="Blender files from folder",
        items= blend_files_end_callback
        )


    all_mesh_blend_files: bpy.props.EnumProperty(
        name="",
        description="Blender files from folder",
        items= all_blend_files_callback
        )


def remove_handy_curve_profile(coll_name):
    profile_coll = bpy.data.collections.get(coll_name)
    if profile_coll:
        bpy.data.collections.remove(profile_coll)
        bpy.ops.outliner.orphans_purge()
        bpy.ops.outliner.orphans_purge()
        bpy.ops.outliner.orphans_purge()

    return


def import_error_happened(self, context):
    self.layout.label(text="Some went wrong with custom file path or Blender file, Please double check.")


def mark_mesh_profile(obj,index):
    obj['HCP_Pro'] = True
    obj['HCP_Ind'] = index
    return

def mark_mesh_end(obj,index,type):
    obj['HCP_End'] = True
    obj['HCP_Ind'] = index
    obj['HCP_Type'] = type
    return

def clear_mesh(obj):
    if obj.get("HCP_Pro"):
        del obj["HCP_Pro"]
    if obj.get("HCP_End"):
        del obj["HCP_End"]
    if obj.get("HCP_Ind"):
        del obj["HCP_Ind"]
    if obj.get("HCP_Type"):
        del obj["HCP_Type"]
    return

def import_handy_curve_profile(coll_name):
    handy_curve_profile_props = bpy.context.scene.handy_curve_profile_props
    bpy.ops.outliner.orphans_purge()
    bpy.ops.outliner.orphans_purge()
    remove_coll = False
   
    if handy_curve_profile_props.profiles_updated:
        remove_coll = True

    if not bpy.data.scenes.get("Handy Curve Profile"):
        remove_coll = True

    else:
        if len(bpy.data.scenes.get("Handy Curve Profile").collection.children) == 0:
            remove_coll = True

        else:
            for child in bpy.data.scenes["Handy Curve Profile"].collection.children:
                if child.name == coll_name:
                    if len(child.objects) == 0:
                        remove_coll = True
                
    
    if remove_coll:
        coll = bpy.data.collections.get(coll_name)
        if coll:
            if len(coll.objects):
                for obj in coll.objects:
                    bpy.data.objects.remove(obj, do_unlink = True)
            bpy.data.collections.remove(coll)

        bpy.ops.outliner.orphans_purge()


    if not bpy.data.collections.get(coll_name):
        pref_path = bpy.context.preferences.addons[__package__].preferences.filepath
        directory = ''
        if pref_path == '':
            script_file = os.path.realpath(__file__)
            directory = os.path.dirname(script_file)
        else:
            directory = pref_path

        blend_file = coll_name + '.blend'
        filepath = os.path.join(directory , blend_file)  


        try:
            if not bpy.data.scenes.get('Handy Curve Profile'):
                bpy.data.scenes.new("Handy Curve Profile")

            with bpy.data.libraries.load(filepath, link=False) as (data_from, data_to):
                data_to.collections = [
                    c for c in data_from.collections if c.startswith(coll_name)]

            for coll in data_to.collections:
                if coll:
                    bpy.data.scenes["Handy Curve Profile"].collection.children.link(coll)

            set_profiles_updated(handy_curve_profile_props,False)

        except:
            bpy.context.window_manager.popup_menu(import_error_happened, title="Error", icon='ERROR')
    return



class ApplyCurveProfileOperator(bpy.types.Operator):
    """ Apply Curve Profile to Curves, Select Curve As Active Object to Start"""
    bl_idname = "hcp_apply_curve_profile.operator"
    bl_label = "Apply Curve Profile"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        if len(context.selected_objects) > 0  and context.active_object.type == 'CURVE':
            c =True
        return  c

    def execute(self, context):
        handy_curve_profile_props = context.scene.handy_curve_profile_props
        objs = context.selected_objects

        try:
            for obj in objs:
                if obj.type != 'CURVE':
                    obj.select_set(False)
                else:
                    if len(obj.data.splines) > 1:
                        bpy.ops.object.mode_set(mode='EDIT')
                        bpy.ops.curve.select_all(action='DESELECT')

                        while len(obj.data.splines) > 1:
                            if obj.data.splines[0].type == 'NURBS' or obj.data.splines[0].type == 'POLY':
                                for point in obj.data.splines[0].points:
                                    point.select = True
                                bpy.ops.curve.separate()
                            elif obj.data.splines[0].type == 'BEZIER':
                                for point in obj.data.splines[0].bezier_points:
                                    point.select_control_point = True
                                bpy.ops.curve.separate()

            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

            curves = context.selected_objects

            profile_mesh, profile_mesh_ext = os.path.splitext(handy_curve_profile_props.profile_mesh_blend_files)

            if not bpy.data.collections.get(profile_mesh) or handy_curve_profile_props.profiles_updated:
                remove_handy_curve_profile(profile_mesh)
                import_handy_curve_profile(profile_mesh)
            Handy_Curve_Profile_Collection = bpy.data.collections[profile_mesh]

            for c in curves:
                curve_profile_mesh = bpy.data.objects.new(c.name + '_profile', Handy_Curve_Profile_Collection.objects[0].data)
                curve_profile_mesh.data = curve_profile_mesh.data.copy()
                c.users_collection[0].objects.link(curve_profile_mesh)

                curve_profile_mesh.matrix_world = c.matrix_world
                array_mod = curve_profile_mesh.modifiers.new("HCP_Array", 'ARRAY')
                array_mod.show_expanded = False
                array_mod.curve = c
                array_mod.count = 3
                array_mod.use_merge_vertices = True
                array_mod.merge_threshold = 0.001
                array_mod.use_merge_vertices_cap = True

                if handy_curve_profile_props.array_type_fit:
                    array_mod.fit_type = 'FIT_CURVE'
                elif handy_curve_profile_props.array_type_manual:
                    array_mod.fit_type = 'FIXED_COUNT'

                cur_mod = curve_profile_mesh.modifiers.new("HCP_Curve", 'CURVE')
                cur_mod.show_expanded = False
                cur_mod.object = c

                c.select_set(False)
                curve_profile_mesh.select_set(True)
                bpy.context.view_layer.objects.active = curve_profile_mesh

                mark_mesh_profile(curve_profile_mesh,1)

        except:
            bpy.context.window_manager.popup_menu(import_error_happened, title="Error", icon='ERROR')
        return {'FINISHED'}


class LoadPrevCurveProfileOperator(bpy.types.Operator):
    """Load Previous Curve Profile in Selected Category """
    bl_idname = "hcp_load_prev_curve_profile.operator"
    bl_label = "Load Prev Curve Profile"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.get('HCP_Pro') :
                c =True
        return  c

    def execute(self, context):
        handy_curve_profile_props = context.scene.handy_curve_profile_props
        profile_mesh, profile_mesh_ext = os.path.splitext(handy_curve_profile_props.profile_mesh_blend_files)
        obj = context.active_object
        
        try:
            if not bpy.data.collections.get(profile_mesh) or handy_curve_profile_props.profiles_updated:
                import_handy_curve_profile(profile_mesh)

            Handy_Curve_Profile_Collection = bpy.data.collections[profile_mesh]
            pos = 1
            if obj.get('HCP_Ind'):
                pos = obj['HCP_Ind']
            else:
                obj.data = Handy_Curve_Profile_Collection.objects[0].data
                obj.data = obj.data.copy()
                mark_mesh_profile(obj,1)
                return {'FINISHED'}

            if pos - 1 < 1:
                pos = len(Handy_Curve_Profile_Collection.objects)
            elif pos - 1 > len(Handy_Curve_Profile_Collection.objects):
                pos = 1
            else:
                pos = pos -1
            
            idx = pos - 1

            obj.data = Handy_Curve_Profile_Collection.objects[idx].data
            obj.data = obj.data.copy()
            mark_mesh_profile(obj,pos)

            bpy.ops.outliner.orphans_purge()

        except:
            bpy.context.window_manager.popup_menu(import_error_happened, title="Error", icon='ERROR')
        return {'FINISHED'}


class LoadNextCurveProfileOperator(bpy.types.Operator):
    """Load Next Curve Profile in Selected Category"""
    bl_idname = "hcp_load_next_curve_profile.operator"
    bl_label = "Load Next Curve Profile"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.get('HCP_Pro') :
                c =True
        return  c

    def execute(self, context):
        handy_curve_profile_props = context.scene.handy_curve_profile_props
        profile_mesh, profile_mesh_ext = os.path.splitext(handy_curve_profile_props.profile_mesh_blend_files)
        obj = context.active_object
        
        try:
            if not bpy.data.collections.get(profile_mesh) or handy_curve_profile_props.profiles_updated:
                import_handy_curve_profile(profile_mesh)

            Handy_Curve_Profile_Collection = bpy.data.collections[profile_mesh]
            pos = 1
            if obj.get('HCP_Ind'):
                pos = obj['HCP_Ind']
            else:
                obj.data = Handy_Curve_Profile_Collection.objects[0].data
                obj.data = obj.data.copy()
                mark_mesh_profile(obj,1)
                return {'FINISHED'}

            if pos + 1 > len(Handy_Curve_Profile_Collection.objects):
                pos = 1

            elif pos + 1 < 1:
                pos = 1

            else:
                pos = pos +1
            
            idx = pos - 1

            obj.data = Handy_Curve_Profile_Collection.objects[idx].data
            obj.data = obj.data.copy()
            mark_mesh_profile(obj,pos)

            bpy.ops.outliner.orphans_purge()
        except:
            bpy.context.window_manager.popup_menu(import_error_happened, title="Error", icon='ERROR')
        return {'FINISHED'}



def parent_objects(child, parent):
    bpy.ops.object.select_all(action='DESELECT')
    child.select_set(True)
    parent.select_set(True)
    bpy.context.view_layer.objects.active = parent
    bpy.ops.object.parent_set(type='OBJECT', keep_transform=False)

    return

class ApplyCurveEndOperator(bpy.types.Operator):
    """ Apply Soft Curve End to Curves, Select Curve As Active Object to Start"""
    bl_idname = "hcp_apply_curve_end.operator"
    bl_label = "Apply Curve End"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        if len(context.selected_objects) > 0  and context.active_object.type == 'CURVE':
            c =True
        return  c

    def execute(self, context):
        handy_curve_profile_props = context.scene.handy_curve_profile_props
        objs = context.selected_objects

        try:
            for obj in objs:
                if obj.type != 'CURVE':
                    obj.select_set(False)
                else:
                    if len(obj.data.splines) > 1:
                        bpy.ops.object.mode_set(mode='EDIT')
                        bpy.ops.curve.select_all(action='DESELECT')

                        while len(obj.data.splines) > 1:
                            if obj.data.splines[0].type == 'NURBS' or obj.data.splines[0].type == 'POLY':
                                for point in obj.data.splines[0].points:
                                    point.select = True
                                bpy.ops.curve.separate()
                            elif obj.data.splines[0].type == 'BEZIER':
                                for point in obj.data.splines[0].bezier_points:
                                    point.select_control_point = True
                                bpy.ops.curve.separate()

            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

            curves = context.selected_objects

            end_mesh, end_mesh_ext = os.path.splitext(handy_curve_profile_props.end_mesh_blend_files)

            if not bpy.data.collections.get(end_mesh) or handy_curve_profile_props.profiles_updated:
                remove_handy_curve_profile(end_mesh)
                import_handy_curve_profile(end_mesh)
            Handy_Curve_Profile_Collection = bpy.data.collections[end_mesh]

            for c in curves:
                curve_end_mesh = bpy.data.objects.new(c.name + '_end', Handy_Curve_Profile_Collection.objects[0].data)
                curve_end_mesh.data = curve_end_mesh.data.copy()
                c.users_collection[0].objects.link(curve_end_mesh)

                cur_proxy_cur = bpy.data.curves.new('proxy' + c.name ,c.type)
                cur_proxy_obj = bpy.data.objects.new('proxy' + c.name , cur_proxy_cur)

                cur_proxy_obj.data =c.data
                cur_proxy_obj.data = cur_proxy_obj.data.copy()
                c.users_collection[0].objects.link(cur_proxy_obj)

                curve_end_mesh.matrix_world = c.matrix_world
                cur_proxy_obj.matrix_world = c.matrix_world

                parent_objects(cur_proxy_obj,c)
                cur_proxy_obj.hide_viewport = True

                cur_mod = curve_end_mesh.modifiers.new("HCP_Curve" , 'CURVE')
                cur_mod.show_expanded = False
                cur_mod.object = cur_proxy_obj

                c.select_set(False)
                curve_end_mesh.select_set(True)
                context.view_layer.objects.active = curve_end_mesh

                mark_mesh_end(curve_end_mesh,1,1)

        except:
            bpy.context.window_manager.popup_menu(import_error_happened, title="Error", icon='ERROR')
        return {'FINISHED'}



class ApplyRigidEndOperator(bpy.types.Operator):
    """Apply Rigid Curve End to Curves, Select Curve As Active Object to Start"""
    bl_idname = "hcp_apply_rigid_end.operator"
    bl_label = "Apply Rigid Curve End to active Curve"
    bl_options = {'REGISTER', 'UNDO'}

    duplicated_obj = None
    cur_obj = None
    proxy = None

    @classmethod
    def poll(cls, context):
        c = False
        if len(context.selected_objects) > 0  and context.active_object.type == 'CURVE':
            c =True
        return  c

    def execute(self, context):
        handy_curve_profile_props = context.scene.handy_curve_profile_props
        objs = context.selected_objects

        try:
            for obj in objs:
                if obj.type != 'CURVE':
                    obj.select_set(False)
                else:
                    if len(obj.data.splines) > 1:
                        bpy.ops.object.mode_set(mode='EDIT')
                        bpy.ops.curve.select_all(action='DESELECT')

                        while len(obj.data.splines) > 1:
                            if obj.data.splines[0].type == 'NURBS' or obj.data.splines[0].type == 'POLY':
                                for point in obj.data.splines[0].points:
                                    point.select = True
                                bpy.ops.curve.separate()
                            elif obj.data.splines[0].type == 'BEZIER':
                                for point in obj.data.splines[0].bezier_points:
                                    point.select_control_point = True
                                bpy.ops.curve.separate()

            bpy.ops.object.mode_set(mode='OBJECT')

            curves = context.selected_objects
            bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

            end_mesh, end_mesh_ext = os.path.splitext(handy_curve_profile_props.end_mesh_blend_files)

            if not bpy.data.collections.get(end_mesh) or handy_curve_profile_props.profiles_updated:
                remove_handy_curve_profile(end_mesh)
                import_handy_curve_profile(end_mesh)
            Handy_Curve_Profile_Collection = bpy.data.collections[end_mesh]

            for c in curves:
                curve_end_mesh = bpy.data.objects.new(c.name + '_end', Handy_Curve_Profile_Collection.objects[0].data)

                curve_end_mesh.data = curve_end_mesh.data.copy()
                c.users_collection[0].objects.link(curve_end_mesh)

                cur_proxy_cur = bpy.data.curves.new('proxy' + c.name , c.type)
                cur_proxy_obj = bpy.data.objects.new('proxy' + c.name , cur_proxy_cur)

                cur_proxy_obj.data =c.data
                cur_proxy_obj.data = cur_proxy_obj.data.copy()
                c.users_collection[0].objects.link(cur_proxy_obj)

                curve_end_mesh.matrix_world = c.matrix_world
                cur_proxy_obj.matrix_world = c.matrix_world

                parent_objects(cur_proxy_obj,c)
                cur_proxy_obj.hide_viewport = True

                bpy.ops.mesh.primitive_plane_add(size=0.001)
                mesh_proxy = context.active_object
                mesh_proxy.name = curve_end_mesh.name
                mesh_proxy.data.name = mesh_proxy.name
                mesh_proxy.matrix_world = c.matrix_world

                curve_modifier = mesh_proxy.modifiers.new(
                    name="HCP_Curve", type='CURVE')
                curve_modifier.object = cur_proxy_obj
                curve_modifier.show_in_editmode = True
                curve_modifier.show_on_cage = True
                curve_modifier.show_expanded = False

                curve_end_mesh.parent = mesh_proxy
                curve_end_mesh.hide_set(True)
                curve_end_mesh.location = (0,0,0)
                mesh_proxy.instance_type = 'FACES'

                mark_mesh_end(mesh_proxy,1,2)

        except:
            bpy.context.window_manager.popup_menu(import_error_happened, title="Error", icon='ERROR')

        return {'FINISHED'}




class LoadPrevCurveEndOperator(bpy.types.Operator):
    """Load Previous Curve End in Selected Category """
    bl_idname = "hcp_load_prev_curve_end.operator"
    bl_label = "Load Prev Curve End"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.get('HCP_End') :
                c =True
        return  c

    def execute(self, context):
        handy_curve_profile_props = context.scene.handy_curve_profile_props
        obj = context.active_object

        end_mesh, end_mesh_ext = os.path.splitext(handy_curve_profile_props.end_mesh_blend_files)

        try:
            if not bpy.data.collections.get(end_mesh) or handy_curve_profile_props.profiles_updated:
                remove_handy_curve_profile(end_mesh)
                import_handy_curve_profile(end_mesh)
            Handy_Curve_Profile_Collection = bpy.data.collections[end_mesh]

            pos = 1
            if obj.get('HCP_Ind'):
                pos = obj['HCP_Ind']
            else:
                obj.data = Handy_Curve_Profile_Collection.objects[0].data
                obj.data = obj.data.copy()
                if obj.get('HCP_Type'):
                    if obj['HCP_Type'] == 1:
                        mark_mesh_end(obj,1,1)
                    elif obj['HCP_Type'] == 2:
                        mark_mesh_end(obj,1,2)
                return {'FINISHED'}

            if obj.get('HCP_Type'):
                if pos - 1 < 1:
                    pos = len(Handy_Curve_Profile_Collection.objects)
                elif pos - 1 > len(Handy_Curve_Profile_Collection.objects):
                    pos = 1
                else:
                    pos = pos -1

                idx = pos -1

                if obj['HCP_Type'] == 1:
                    
                    obj.data = Handy_Curve_Profile_Collection.objects[idx].data
                    obj.data = obj.data.copy()

                    mark_mesh_end(obj,pos,1)

                elif obj['HCP_Type'] == 2:
                    actual_mesh = obj.children[0]
                    actual_mesh.data = Handy_Curve_Profile_Collection.objects[idx].data
                    actual_mesh.data = actual_mesh.data.copy()

                    mark_mesh_end(obj,pos,2)

            bpy.ops.outliner.orphans_purge()
        except:
            bpy.context.window_manager.popup_menu(import_error_happened, title="Error", icon='ERROR')
        return {'FINISHED'}


class LoadNextCurveEndOperator(bpy.types.Operator):
    """Load Next Curve End in Selected Category"""
    bl_idname = "hcp_load_next_curve_end.operator"
    bl_label = "Load Next Curve End"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.get('HCP_End') :
                c =True
        return  c


    def execute(self, context):
        handy_curve_profile_props = context.scene.handy_curve_profile_props
        obj = context.active_object
        end_mesh, end_mesh_ext = os.path.splitext(handy_curve_profile_props.end_mesh_blend_files)

        try:
            if not bpy.data.collections.get(end_mesh) or handy_curve_profile_props.profiles_updated:
                remove_handy_curve_profile(end_mesh)
                import_handy_curve_profile(end_mesh)
            Handy_Curve_Profile_Collection = bpy.data.collections[end_mesh]

            pos = 1
            if obj.get('HCP_Ind'):
                pos = obj['HCP_Ind']
            else:
                obj.data = Handy_Curve_Profile_Collection.objects[0].data
                obj.data = obj.data.copy()
                if obj.get('HCP_Type'):
                    if obj['HCP_Type'] == 1:
                        mark_mesh_end(obj,1,1)
                    elif obj['HCP_Type'] == 2:
                        mark_mesh_end(obj,1,2)
               
                return {'FINISHED'}

            if obj.get('HCP_Type'):
                if pos + 1 > len(Handy_Curve_Profile_Collection.objects):
                    pos = 1
                elif pos + 1 < 1:
                    pos = 1
                else:
                    pos = pos +1

                idx = pos - 1

                if obj['HCP_Type'] == 1:
                    
                    obj.data = Handy_Curve_Profile_Collection.objects[idx].data
                    obj.data = obj.data.copy()

                    mark_mesh_end(obj,pos,1)

                elif obj['HCP_Type'] == 2:
                    actual_mesh = obj.children[0]
                    actual_mesh.data = Handy_Curve_Profile_Collection.objects[idx].data
                    actual_mesh.data = actual_mesh.data.copy()

                    mark_mesh_end(obj,pos,2)

            bpy.ops.outliner.orphans_purge()
        except:
            bpy.context.window_manager.popup_menu(import_error_happened, title="Error", icon='ERROR')
        return {'FINISHED'}




class Test_Mesh_Before_Append_Operator(bpy.types.Operator):
    """Test mesh shape before append to source file"""
    bl_idname = "test_mesh_before_append.operator"
    bl_label = "Test Mesh Before Append Operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        if context.active_object:
            if context.active_object.type == 'MESH':
                c =True
        return  c


    def execute(self, context):
        obj = context.active_object

        if obj.modifiers.get('Test_Array') or obj.modifiers.get('Test_Curve'):
            if obj.modifiers.get('Test_Array'):
                modifier_to_remove = obj.modifiers["Test_Array"]
                obj.modifiers.remove(modifier_to_remove)
            if obj.modifiers.get('Test_Curve'):
                modifier_to_remove = obj.modifiers["Test_Curve"]
                cur = modifier_to_remove.object
                obj.modifiers.remove(modifier_to_remove)
                bpy.data.objects.remove(cur, do_unlink = True)
                bpy.ops.outliner.orphans_purge()
        else:
            bpy.ops.curve.primitive_nurbs_path_add()
            cur = context.active_object
            cur.matrix_world = obj.matrix_world

            array_mod = obj.modifiers.new("Test_Array", 'ARRAY')
            array_mod.show_expanded = False
            array_mod.curve = cur
            array_mod.count = 3
            array_mod.use_merge_vertices = True
            array_mod.merge_threshold = 0.001
            array_mod.use_merge_vertices_cap = True
            array_mod.show_in_editmode = True
            array_mod.show_on_cage = True

            cur_mod = obj.modifiers.new("Test_Curve", 'CURVE')
            cur_mod.show_expanded = False
            cur_mod.object = cur
            cur_mod.show_in_editmode = True
            cur_mod.show_on_cage = True

            cur.select_set(False)
            obj.select_set(True)
            context.view_layer.objects.active = obj

        return {'FINISHED'}


def not_saved_warning(self, context):
    self.layout.label(text="Save Blend File First.")

class HCP_Append_to_Source_Operator(bpy.types.Operator):
    """Append Selected Mesh to Source Blend File For Future Usage"""
    bl_idname = "hcp_append_to_source.operator"
    bl_label = "Append_to_Source"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        if len(objs)==1 :
            if context.active_object == objs[0] and objs[0].type == 'MESH' and bpy.path.basename(bpy.context.blend_data.filepath) != bpy.context.scene.handy_curve_profile_props.all_mesh_blend_files:
                c =True
        return  c

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)


    def execute(self, context):
        handy_curve_profile_props = bpy.context.scene.handy_curve_profile_props
        blend_file_name, blend_file_ext = os.path.splitext(handy_curve_profile_props.all_mesh_blend_files)
        coll_name = blend_file_name
        obj = bpy.context.active_object
        pref_path = context.preferences.addons[__package__].preferences.filepath
        
        if obj.mode != "OBJECT":
            bpy.ops.object.mode_set(mode='OBJECT')

        if not bpy.data.is_saved:
            context.window_manager.popup_menu(not_saved_warning, title="Error", icon='ERROR')
        else:
            try:
                bpy.ops.wm.save_mainfile()

                bpy.ops.object.parent_clear(type='CLEAR_KEEP_TRANSFORM')
                bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

                obj.name = 'HCP_New_Temp_Mesh'
                
                if len(obj.children):
                    for ob in obj.children:
                        bpy.data.objects.remove(ob, do_unlink=True)
                
                bpy.ops.view3d.copybuffer()
                source_directory = ''
                if pref_path == '':
                    script_file = os.path.realpath(__file__)
                    source_directory = os.path.dirname(script_file)
                else:
                    source_directory = pref_path

                source_file_path = os.path.join(source_directory , handy_curve_profile_props.all_mesh_blend_files)  
                bpy.ops.wm.open_mainfile(filepath= source_file_path)

                bpy.ops.view3d.pastebuffer()
                new_mesh = bpy.data.objects['HCP_New_Temp_Mesh']
                new_mesh.name = new_mesh.name.replace("HCP_New_Temp_", "")
                
                if not bpy.data.collections.get(coll_name):
                    collection = bpy.context.blend_data.collections.new(name=coll_name)
                    bpy.context.collection.children.link(collection)

                target_collection = bpy.data.collections[coll_name]
                
                new_mesh.users_collection[0].objects.unlink(new_mesh)

                target_collection.objects.link(new_mesh)

                bpy.ops.wm.save_mainfile()

                if bpy.app.version[0] ==2:
                    fp = bpy.utils.user_resource('CONFIG', "recent-files.txt")
                elif bpy.app.version[0] >=3:
                    fp = bpy.utils.user_resource('CONFIG',  path="recent-files.txt")

                try:
                    with open(fp) as file:
                        recent_files = file.read().splitlines()
                except (IOError, OSError, FileNotFoundError):
                    recent_files = []
                    
                bpy.ops.wm.open_mainfile(filepath= recent_files[0])


                handy_curve_profile_props = bpy.context.scene.handy_curve_profile_props
                set_profiles_updated(handy_curve_profile_props,True)

            except:
                bpy.context.window_manager.popup_menu(import_error_happened, title="Error", icon='ERROR')


        return {'FINISHED'}

class UpdateProxyCurveOperator(bpy.types.Operator):
    """Update When Curve is Modified"""
    bl_idname = "hcp_update_proxy_curve.operator"
    bl_label = "Update Curve Modification"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.get('HCP_End'):
                c =True
        return  c

    def execute(self, context):
        obj = context.active_object
        cur_mod = None
        cur_proxy = None
        cur_proxy_new = None
        cur_master = None
        for mod in obj.modifiers:
            if mod.name == 'HCP_Curve':
                cur_mod = mod
        if cur_mod:
            cur_proxy = cur_mod.object
        if cur_proxy:
            cur_master = cur_proxy.parent
        if cur_master:
            bpy.data.objects.remove(cur_proxy, do_unlink = True)
            bpy.ops.outliner.orphans_purge()

            cur_proxy_cur = bpy.data.curves.new('proxy' + cur_master.name  ,cur_master.type)
            cur_proxy_new = bpy.data.objects.new( 'proxy' + cur_master.name, cur_proxy_cur)
            cur_proxy_new.data =cur_master.data
            cur_proxy_new.data = cur_proxy_new.data.copy()
            cur_master.users_collection[0].objects.link(cur_proxy_new)
            cur_proxy_new.matrix_world = cur_master.matrix_world
            parent_objects(cur_proxy_new,cur_master)

            cur_mod.object = cur_proxy_new
            cur_proxy_new.hide_viewport = True
            cur_master.select_set(False)

        obj.select_set(True)
        bpy.context.view_layer.objects.active = obj

        return {'FINISHED'}


def switch_curve_end(mesh_at_curve_end):
    obj = mesh_at_curve_end
    cur_mod = None
    cur_proxy = None
    for mod in obj.modifiers:
        if mod.name == 'HCP_Curve':
            cur_mod = mod
    if cur_mod:
        cur_proxy = cur_mod.object
    if cur_proxy :
        cur_proxy.hide_viewport = False
        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.ops.object.select_all(action='DESELECT')
        cur_proxy.select_set(True)
        bpy.context.view_layer.objects.active = cur_proxy
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.curve.select_all(action='SELECT')
        bpy.ops.curve.switch_direction()
        bpy.ops.object.mode_set(mode='OBJECT')
        cur_proxy.hide_viewport = True
    obj.select_set(True)
    bpy.context.view_layer.objects.active = obj

    return

class SwitchCurveEndOperator(bpy.types.Operator):
    """Flip to Other End of Curve"""
    bl_idname = "hcp_switch_curve_end.operator"
    bl_label = "Switch Curve End"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.get('HCP_End'):
                c =True
        return  c

    def execute(self, context):
        obj = context.active_object

        switch_curve_end(obj)
        return {'FINISHED'}

class ApplyBothEndOperator(bpy.types.Operator):
    """Duplicate Selected To Both End of Curve"""
    bl_idname = "hcp_apply_both_curve_end.operator"
    bl_label = "Duplicate Selected To Both End of Curve"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.get('HCP_End'):
                c =True
        return  c

    def execute(self, context):
        obj = context.active_object
        c_proxy = None

        if obj.modifiers.get('HCP_Curve'):
            c_proxy = obj.modifiers['HCP_Curve'].object
        if c_proxy and obj.get('HCP_Type'):
            if obj['HCP_Type']==1:
                c_proxy.hide_viewport = False
                bpy.ops.object.select_all(action='DESELECT')
                c_proxy.select_set(True)
                obj.select_set(True)
                bpy.context.view_layer.objects.active = obj
                bpy.ops.object.duplicate()

                for o in c_proxy.parent.children:
                    if 'proxy' in o.name :
                        o.hide_viewport = True

                switch_curve_end(obj)

            elif obj['HCP_Type']==2:
                actual_mesh = obj.children[0]
                actual_mesh.hide_set(False)
                c_proxy.hide_viewport = False
                bpy.ops.object.select_all(action='DESELECT')
                c_proxy.select_set(True)
                obj.select_set(True)
                actual_mesh.select_set(True)

                bpy.context.view_layer.objects.active = obj
                bpy.ops.object.duplicate()
                new_end_mesh = context.active_object
                new_end_mesh.children[0].hide_set(True)
                actual_mesh.hide_set(True)
                for o in c_proxy.parent.children:
                    if 'proxy' in o.name :
                        o.hide_viewport = True
                switch_curve_end(obj)

        return {'FINISHED'}

class FinalizeSelectionOperator(bpy.types.Operator):
    """Finalize the Selected Mesh Profile, Apply Modifiers and Clean Up Proxy """
    bl_idname = "hcp_finalize_selection.operator"
    bl_label = "Finalize the Selection and Apply Modifiers"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj.type == 'MESH': 
            if obj.get('HCP_Pro') or obj.get('HCP_End'):
                c =True
        return  c

    def execute(self, context):
        objs = context.selected_objects

        conv_mesh_list = []
        make_real_list = []
        del_cur_list = []

        for obj in objs:
            if obj.get('HCP_Pro'):
                conv_mesh_list.append(obj)
            elif obj.get('HCP_End') and obj.get('HCP_Type'):
                if obj['HCP_Type'] ==1:
                    conv_mesh_list.append(obj)
                    if obj.modifiers.get('HCP_Curve'):
                        cur = obj.modifiers['HCP_Curve'].object
                        del_cur_list.append(cur)
                elif obj['HCP_Type'] ==2:
                    make_real_list.append(obj)
                    if obj.modifiers.get('HCP_Curve'):
                        cur = obj.modifiers['HCP_Curve'].object
                        del_cur_list.append(cur)
            obj.select_set(False)
        
        if len(conv_mesh_list):
            for obj in conv_mesh_list:
                obj.select_set(True)
                clear_mesh(obj)
                bpy.context.view_layer.objects.active = obj
            
            bpy.ops.object.convert(target='MESH')

        bpy.ops.object.select_all(action='DESELECT')

        if len(make_real_list):
            for obj in make_real_list:
                obj.select_set(True)
                bpy.context.view_layer.objects.active = obj
            bpy.ops.object.duplicates_make_real()
        
        if len(del_cur_list):
            for obj in del_cur_list:
                bpy.data.objects.remove(obj, do_unlink = True)
        
        if len(make_real_list):
            for obj in make_real_list:
                for ob in obj.children:
                    bpy.data.objects.remove(ob, do_unlink = True)
                bpy.data.objects.remove(obj, do_unlink = True)

        bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')
        bpy.ops.outliner.orphans_purge()
        bpy.ops.object.select_all(action='DESELECT')
        
        return {'FINISHED'}

class HANDY_CURVE:
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "HCP"


class HANDY_CURVE_PT_main(HANDY_CURVE, Panel):
    bl_idname = "HANDY_CURVE_PT_main"
    bl_label = "Handy Curve Profile"
    # bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj is not None :
            c =True
        return  c

    def draw(self, context):
        layout = self.layout

        handy_curve_profile_props = context.scene.handy_curve_profile_props
        obj = context.active_object
        array_mod = None
        cur = None
        for mod in obj.modifiers:
            if mod.name == 'HCP_Array':
                array_mod = mod
            elif mod.name == 'HCP_Curve':
                cur = mod.object

        box = layout.box()
        row = box.row()
        row.label(text='Profile Category' )
        row.prop(handy_curve_profile_props, 'profile_mesh_blend_files',
                 text='')
        row = box.row()
        row.scale_y = 1.2
        row.operator("hcp_apply_curve_profile.operator",
                     text='Apply Curve Profile')
        row = box.row()
        row.scale_y = 1.5
        row.operator("hcp_load_prev_curve_profile.operator",
                     text='Prev')
        row.operator("hcp_load_next_curve_profile.operator",
                     text='Next')

        box = layout.box()
        row = box.row()
        row.label(text='Pick an End' )
        row.prop(handy_curve_profile_props, 'end_mesh_blend_files',
                 text='')
        row = box.row()
        row.scale_y = 1.2
        row.operator("hcp_apply_curve_end.operator",
                     text='Apply Soft End')
        row.operator("hcp_apply_rigid_end.operator",
                     text='Apply Rigid End')
        row = box.row()
        row.scale_y = 1.5
        row.operator("hcp_load_prev_curve_end.operator",
                     text='Prev')
        row.operator("hcp_load_next_curve_end.operator",
                     text='Next')
        row = box.row()
        row.scale_y = 1.2
        row.operator("hcp_update_proxy_curve.operator",
                     text='Update')

        row.operator("hcp_switch_curve_end.operator",
                     text='Flip End')
        row.operator("hcp_apply_both_curve_end.operator",
                     text='Both End')

        if obj.get('HCP_Pro'):
            box = layout.box()
            box.label(text='Adjustment' )
            row = box.row()
            row.scale_y = 1.5
            row.prop(handy_curve_profile_props, 'array_type_manual', text='Manual', toggle=True)
            row.prop(handy_curve_profile_props, 'array_type_fit', text='Fit Curve', toggle=True)

            box.prop(handy_curve_profile_props, 'curve_profile_scale',text='Scale', slider=True)
            row = box.row()
            row.prop(handy_curve_profile_props, 'curve_profile_squash',text='Squash', slider=True)
            row.prop(handy_curve_profile_props, 'curve_profile_stretch',text='Stretch', slider=True)

            if array_mod:
                if array_mod.fit_type == 'FIXED_COUNT':
                    box.prop(handy_curve_profile_props, 'segment_count',text='Segment Count')
                box.prop(handy_curve_profile_props, 'segment_distance',text='Segment Distance')

            row = box.row()
            row.prop(handy_curve_profile_props, 'curve_profile_offset',text='Offset', slider=True)
            row.prop(handy_curve_profile_props, 'curve_profile_rotation', text='Rotation', slider=True)

            layout.separator(factor=1.5)
            layout.prop(handy_curve_profile_props, 'radius_tilt', text='Radius and Tilt', toggle=True)
            if handy_curve_profile_props.radius_tilt:
                box = layout.box()
                row = box.row()
                row.scale_y = 1.2
                row.operator("hcp_reset_cur_radius.operator",
                                text='Reset Radius')
                row.operator("hcp_reset_cur_tilt.operator",
                                text='Reset Tilt')
                box = layout.box()

                if cur:
                    if cur.data.splines[0].type == 'BEZIER':
                        for pt in cur.data.splines[0].bezier_points:
                            row = box.row()
                            row.prop(pt, 'radius',text='Radius')
                            row.prop(pt, 'tilt',text='Tilt')
                    elif cur.data.splines[0].type == 'NURBS':
                        for pt in cur.data.splines[0].points:
                            row = box.row()
                            row.prop(pt, 'radius',text='Radius')
                            row.prop(pt, 'tilt',text='Tilt')
                    elif cur.data.splines[0].type == 'POLY':
                        for pt in cur.data.splines[0].points:
                            row = box.row()
                            row.prop(pt, 'radius',text='Radius')
                            row.prop(pt, 'tilt',text='Tilt')

            box = layout.box()
            row = box.row()
            row.scale_y = 1.5
            row.operator("hcp_finalize_selection.operator",text='Finalize Selection')

        elif obj.get('HCP_End') and obj.get('HCP_Type'):
            if obj['HCP_Type'] == 1:
                box = layout.box()
                box.label(text='Adjustment' )

                box.prop(handy_curve_profile_props, 'curve_profile_scale',text='Scale', slider=True)
                row = box.row()
                row.prop(handy_curve_profile_props, 'curve_profile_squash',text='Squash', slider=True)
                row.prop(handy_curve_profile_props, 'curve_profile_stretch',text='Stretch', slider=True)

                row = box.row()
                row.prop(handy_curve_profile_props, 'curve_profile_offset',text='Offset', slider=True)
                row.prop(handy_curve_profile_props, 'curve_profile_rotation', text='Rotation', slider=True)

                layout.separator(factor=1.5)
                layout.prop(handy_curve_profile_props, 'radius_tilt', text='Radius and Tilt', toggle=True)
                if handy_curve_profile_props.radius_tilt:
                    box = layout.box()
                    row = box.row()
                    row.scale_y = 1.2
                    row.operator("hcp_reset_cur_radius.operator",
                                    text='Reset Radius')
                    row.operator("hcp_reset_cur_tilt.operator",
                                    text='Reset Tilt')
                    box = layout.box()

                    if cur:
                        if cur.data.splines[0].type == 'BEZIER':
                            for pt in cur.data.splines[0].bezier_points:
                                row = box.row()
                                row.prop(pt, 'radius',text='Radius')
                                row.prop(pt, 'tilt',text='Tilt')
                        elif cur.data.splines[0].type == 'NURBS':
                            for pt in cur.data.splines[0].points:
                                row = box.row()
                                row.prop(pt, 'radius',text='Radius')
                                row.prop(pt, 'tilt',text='Tilt')
                        elif cur.data.splines[0].type == 'POLY':
                            for pt in cur.data.splines[0].points:
                                row = box.row()
                                row.prop(pt, 'radius',text='Radius')
                                row.prop(pt, 'tilt',text='Tilt')

                box = layout.box()
                row = box.row()
                row.scale_y = 1.5
                row.operator("hcp_finalize_selection.operator",text='Finalize Selection')

            elif obj['HCP_Type'] == 2:
                box = layout.box()
                box.label(text='Adjustment' )
                box.prop(handy_curve_profile_props, 'curve_profile_scale',text='Scale', slider=True)

                row = box.row()
                row.prop(handy_curve_profile_props, 'curve_profile_offset',text='Offset', slider=True)
                row.prop(handy_curve_profile_props, 'curve_profile_rotation', text='Rotation', slider=True)

                box = layout.box()
                row = box.row()
                row.scale_y = 1.5
                row.operator("hcp_finalize_selection.operator",text='Finalize Selection')

        layout.separator(factor=1.5)
        layout.prop(handy_curve_profile_props, 'misc_toggle', text='MISC', toggle=True)
        if handy_curve_profile_props.misc_toggle:
            box = layout.box()
            row = box.row()
            row.scale_y = 1.5
            sub = row.row(align=True)
            sub.scale_x = 2.0
            sub.operator("hcp_bones_from_curve_points.operator",text='Bones From Curves')
            sub.scale_x = 0.5
            sub.prop(handy_curve_profile_props, 'bone_cyclic_toggle', text='Cyclic', toggle=True)

            row = box.row()
            row.scale_y = 1.5
            row.operator("hcp_copy_curve_profile_to_others.operator",text='Copy Profile As Active')

            row = box.row()
            row.scale_y = 1.5
            row.operator("hcp_edge_to_curve.operator",text='Edge To Curve')

            row = box.row()
            row.scale_y = 1.5
            row.operator("hcp_draw_curve_on_surface.operator",text='Draw Curve On Mesh Surface')


            layout.separator(factor=1)
            box = layout.box()
            row = box.row()
            row.scale_y = 1.5
            if obj.modifiers.get('Test_Array') or obj.modifiers.get('Test_Curve'):
                row.operator("test_mesh_before_append.operator",text='Clear Modifier')
            else:
                row.operator("test_mesh_before_append.operator",text='Test Mesh Before Append')

            row = box.row()
            row.scale_y = 1.5
            sub = row.row(align=True)
            sub.scale_x = 2.0
            sub.operator("hcp_append_to_source.operator",text='Append To Source')
            sub.scale_x = 1
            sub.prop(handy_curve_profile_props, 'all_mesh_blend_files', text='')

class ResetCurRadiusOperator(bpy.types.Operator):
    """Reset Curve Radius"""
    bl_idname = "hcp_reset_cur_radius.operator"
    bl_label = "Reset Curve Radius"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        obj = context.active_object
        cur = None
        if obj.modifiers.get('HCP_Curve'):
            cur = obj.modifiers['HCP_Curve'].object
        if cur:
            if cur.data.splines[0].type == 'BEZIER':
                for pt in cur.data.splines[0].bezier_points:
                    pt.radius = 1
            elif cur.data.splines[0].type == 'NURBS':
                for pt in cur.data.splines[0].points:
                    pt.radius = 1
            elif cur.data.splines[0].type == 'POLY':
                for pt in cur.data.splines[0].points:
                    pt.radius = 1

        return {'FINISHED'}

class ResetCurTiltOperator(bpy.types.Operator):
    """Reset Curve Tilt"""
    bl_idname = "hcp_reset_cur_tilt.operator"
    bl_label = "Reset Curve Tilt"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        obj = context.active_object
        cur = None
        if obj.modifiers.get('HCP_Curve'):
            cur = obj.modifiers['HCP_Curve'].object
        if cur:
            if cur.data.splines[0].type == 'BEZIER':
                for pt in cur.data.splines[0].bezier_points:
                    pt.tilt = 0
            elif cur.data.splines[0].type == 'NURBS':
                for pt in cur.data.splines[0].points:
                    pt.tilt = 0
            elif cur.data.splines[0].type == 'POLY':
                for pt in cur.data.splines[0].points:
                    pt.tilt = 0

        return {'FINISHED'}




class Edge_To_Curve_Operator(bpy.types.Operator):
    """Turn Selected Edges Into Curves"""
    bl_idname = "hcp_edge_to_curve.operator"
    bl_label = "Edge_To_Curve"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj is not None :
            if obj.type == 'MESH' and obj.mode == 'EDIT' :
                c =True
        return  c

    def execute(self, context):
        objs = context.selected_objects
        for ob in objs:
            if ob.type != 'MESH':
                ob.select_set(False)

        obj = context.active_object
        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.ops.object.mode_set(mode='EDIT')
        selected_edges = [e for e in obj.data.edges if e.select]

        if len(selected_edges)>0:
            bpy.ops.mesh.duplicate()
            bpy.ops.mesh.separate(type='SELECTED')
            bpy.ops.object.mode_set(mode='OBJECT')
            objs_new = context.selected_objects

            edge_objs = []
            for ob in objs_new:
                ob.select_set(False)
                if ob not in objs:
                    edge_objs.append(ob)

            for ob in edge_objs:
                ob.select_set(True)

            context.view_layer.objects.active = edge_objs[0]
            bpy.ops.object.convert(target='CURVE')

            for ob in edge_objs:
                ob.data.bevel_depth = 0.003

            bpy.ops.object.mode_set(mode='OBJECT')
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')

        return {'FINISHED'}



class Copy_Curve_Profile_To_Others_Operator(bpy.types.Operator):
    """Copy curve profile from active object to others"""
    bl_idname = "hcp_copy_curve_profile_to_others.operator"
    bl_label = "Copy_Curve_Profile_To_Others"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj is not None :
            if obj.get('HCP_Pro') or obj.get('HCP_End'):
                c =True
        return  c

    def execute(self, context):
        obj = context.active_object
        objs = context.selected_objects

        if obj.get('HCP_Pro'):
            for ob in objs:
                if ob.get('HCP_Pro'):
                    ob.data = obj.data
                    ob.data = ob.data.copy()

        elif obj.get('HCP_End') and obj.get('HCP_Type'):
            if obj['HCP_Type'] ==1:
                for ob in objs:
                    if ob.get('HCP_Type') and ob['HCP_Type'] ==1:
                        ob.data = obj.data
                        ob.data = ob.data.copy()

            elif obj['HCP_Type'] ==2:
                for ob in objs:
                    if ob.get('HCP_Type') and ob['HCP_Type'] ==2:
                        ob.children[0].data = obj.children[0].data
                        ob.children[0].data = ob.children[0].data.copy()

        bpy.ops.outliner.orphans_purge()

        return {'FINISHED'}



class Bones_From_Curve_Points_OP(Operator):
    """Create Bones From Curve Points, select a curve to start"""
    bl_label = "Create Bones From Curve Points"
    bl_idname = "hcp_bones_from_curve_points.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        if len(objs)>0:
            if context.active_object.type == 'CURVE':
                c = True
        return c

    def execute(self, context):
        handy_curve_profile_props = context.scene.handy_curve_profile_props
        bpy.ops.object.mode_set(mode='OBJECT')
        objs = context.selected_objects
        for obj in objs:
            if obj.type != 'CURVE':
                obj.select_set(False)

        objs = context.selected_objects
        if len(objs) == 0:
            return {'FINISHED'}
        else:
            if not handy_curve_profile_props.bone_cyclic_toggle:
                for obj in objs:
                    cur_data = obj.data
                    arm_data = bpy.data.armatures.new("Armature_" + obj.name)
                    arm_obj = bpy.data.objects.new("Armature_" + obj.name, arm_data)
                    context.scene.collection.objects.link(arm_obj)
                    arm_obj.show_in_front = True
                    context.view_layer.objects.active = arm_obj
                    arm_obj.matrix_world = obj.matrix_world
                    bpy.ops.object.mode_set(mode='EDIT')

                    for spline in cur_data.splines:
                        if spline.type == 'BEZIER':
                            parent = None
                            pt_index = 0
                            for i, bezier_point in enumerate(spline.bezier_points):
                                bone = arm_data.edit_bones.new("Bone")
                                bone.head = spline.bezier_points[i].co
                                bone.tail = spline.bezier_points[i+1].co
                                bone.parent = parent
                                bone.use_connect = True
                                parent = bone
                                pt_index = pt_index + 1
                                if pt_index > len(spline.bezier_points)-2:
                                    break

                        elif spline.type == 'NURBS' or spline.type == 'POLY':
                            parent = None
                            pt_index = 0
                            for i, point in enumerate(spline.points):
                                bone = arm_data.edit_bones.new("Bone")
                                v=[0,0,0]
                                v[0] = spline.points[i].co[0]
                                v[1] = spline.points[i].co[1]
                                v[2] = spline.points[i].co[2]
                                bone.head = v
                                v[0] = spline.points[i+1].co[0]
                                v[1] = spline.points[i+1].co[1]
                                v[2] = spline.points[i+1].co[2]
                                bone.tail = v
                                bone.parent = parent
                                bone.use_connect = True
                                parent = bone
                                pt_index = pt_index + 1
                                if pt_index > len(spline.points)-2:
                                    break

                bpy.ops.object.mode_set(mode='OBJECT')
                bpy.context.view_layer.objects.active = objs[0]

            else:
                for obj in objs:
                    if len(obj.data.splines) > 1:
                        bpy.ops.object.mode_set(mode='EDIT')
                        bpy.ops.curve.select_all(action='DESELECT')

                        while len(obj.data.splines) > 1:
                            if obj.data.splines[0].type == 'NURBS' or obj.data.splines[0].type == 'POLY':
                                for point in obj.data.splines[0].points:
                                    point.select = True
                                bpy.ops.curve.separate()
                            elif obj.data.splines[0].type == 'BEZIER':
                                for point in obj.data.splines[0].bezier_points:
                                    point.select_control_point = True
                                bpy.ops.curve.separate()

                bpy.ops.object.mode_set(mode='OBJECT')

                objs = context.selected_objects

                for obj in objs:
                    cur_data = obj.data
                    arm_data = bpy.data.armatures.new("Armature_" + obj.name)
                    arm_obj = bpy.data.objects.new("Armature_" + obj.name, arm_data)
                    context.scene.collection.objects.link(arm_obj)
                    arm_obj.show_in_front = True
                    context.view_layer.objects.active = arm_obj
                    arm_obj.matrix_world = obj.matrix_world
                    bpy.ops.object.mode_set(mode='EDIT')

                    for spline in cur_data.splines:
                        if spline.type == 'BEZIER':
                            parent = None
                            pt_index = 0
                            for i, bezier_point in enumerate(spline.bezier_points):
                                bone = arm_data.edit_bones.new("Bone")
                                bone.head = spline.bezier_points[i].co
                                bone.tail = spline.bezier_points[i+1].co
                                bone.parent = parent
                                bone.use_connect = True
                                parent = bone
                                pt_index = pt_index + 1
                                if pt_index > len(spline.bezier_points)-2:
                                    break

                            bone = arm_data.edit_bones.new("Bone")
                            bone.tail = spline.bezier_points[0].co
                            bone.head = spline.bezier_points[len(spline.bezier_points)-1].co
                            bone.parent = arm_data.edit_bones[len(arm_data.edit_bones)-2]
                            bone.use_connect = True
                            arm_data.edit_bones[0].parent = bone
                            arm_data.edit_bones[0].use_connect = True

                        elif spline.type == 'NURBS' or spline.type == 'POLY':
                            parent = None
                            pt_index = 0
                            for i, point in enumerate(spline.points):
                                bone = arm_data.edit_bones.new("Bone")
                                v=[0,0,0]
                                v[0] = spline.points[i].co[0]
                                v[1] = spline.points[i].co[1]
                                v[2] = spline.points[i].co[2]
                                bone.head = v
                                v[0] = spline.points[i+1].co[0]
                                v[1] = spline.points[i+1].co[1]
                                v[2] = spline.points[i+1].co[2]
                                bone.tail = v
                                bone.parent = parent
                                bone.use_connect = True
                                parent = bone
                                pt_index = pt_index + 1
                                if pt_index > len(spline.points)-2:
                                    break

                            bone = arm_data.edit_bones.new("Bone")
                            v=[0,0,0]
                            v[0] = spline.points[0].co[0]
                            v[1] = spline.points[0].co[1]
                            v[2] = spline.points[0].co[2]
                            bone.tail = v
                            v[0] = spline.points[len(spline.points)-1].co[0]
                            v[1] = spline.points[len(spline.points)-1].co[1]
                            v[2] = spline.points[len(spline.points)-1].co[2]
                            bone.head = v
                            bone.parent = arm_data.edit_bones[len(arm_data.edit_bones)-2]
                            bone.use_connect = True
                            arm_data.edit_bones[0].parent = bone
                            arm_data.edit_bones[0].use_connect = True

                    bpy.ops.object.mode_set(mode='OBJECT')
                    bpy.context.view_layer.objects.active = objs[0]

        return {'FINISHED'}


class Draw_Curve_On_Surface_OP(Operator):
    """Darw Curves On Mesh Surface, select a mesh to start"""
    bl_label = "Draw_Curve_On_Surface"
    bl_idname = "hcp_draw_curve_on_surface.operator"
    bl_options = {"REGISTER", "UNDO", "BLOCKING"}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        if len(objs)>0:
            if context.active_object.type == 'MESH':
                c = True
        return c


    def prep(self, context):
        self.vert_list = []
        self.init_snap = context.scene.tool_settings.use_snap
        self.init_element = context.scene.tool_settings.snap_elements

        context.scene.tool_settings.use_snap = True
        context.scene.tool_settings.snap_elements = {'FACE'}

        objs = context.selected_objects
        for obj in objs:
            if obj.type != 'MESH':
                obj.select_set(False)
        bpy.context.view_layer.objects.active = objs[0]
        bpy.ops.object.duplicate()
        bpy.ops.object.join()
        bpy.ops.object.hide_view_set(unselected=True)

        self.target_mesh = context.active_object
        bpy.ops.mesh.primitive_plane_add()
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.mesh.select_mode(type="VERT")
        bpy.ops.mesh.select_all(action='SELECT')
        bpy.ops.mesh.delete(type='VERT')
        self.new_obj = context.active_object
        self.new_obj.show_in_front = True
        self.new_obj.name = 'New Curve'

        self.subd_mod = self.new_obj.modifiers.new("subd_mod", 'SUBSURF')
        self.subd_mod.levels = 1
        self.subd_mod.show_on_cage = True
        self.subd_mod.show_expanded = False

        self.wrap_mod = self.new_obj.modifiers.new("wrap_mod", 'SHRINKWRAP')
        self.wrap_mod.target = self.target_mesh
        self.wrap_mod.show_on_cage = True
        self.wrap_mod.wrap_mode = 'ABOVE_SURFACE'
        self.wrap_mod.show_expanded = False



    def finish_up(self, context):
        bpy.ops.object.mode_set(mode='OBJECT')
        context.scene.tool_settings.use_snap = self.init_snap
        context.scene.tool_settings.snap_elements = self.init_element

        bpy.ops.object.modifier_apply( modifier = self.subd_mod.name )
        bpy.ops.object.modifier_apply( modifier = self.wrap_mod.name )

        bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')
        bpy.ops.object.convert(target='CURVE')

        self.new_obj.show_in_front = False

        bpy.data.objects.remove(self.target_mesh, do_unlink = True)

        bpy.ops.object.hide_view_clear()
        bpy.ops.object.select_all(action='DESELECT')
        self.new_obj.select_set(True)
        bpy.context.view_layer.objects.active = self.new_obj

        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.curve.smooth()
        bpy.ops.curve.smooth()

        bpy.ops.curve.select_all(action='SELECT')
        bpy.ops.curve.spline_type_set(type='BEZIER')
        bpy.ops.curve.handle_type_set(type='ALIGNED')
        bpy.ops.curve.decimate(ratio=0.3)

        bpy.ops.object.mode_set(mode='OBJECT')
        context.workspace.status_text_set("")

    def canceled_op(self, context):
        bpy.ops.object.mode_set(mode='OBJECT')
        context.scene.tool_settings.use_snap = self.init_snap
        context.scene.tool_settings.snap_elements = self.init_element

        bpy.data.objects.remove(self.new_obj, do_unlink = True)
        bpy.data.objects.remove(self.target_mesh, do_unlink = True)

        bpy.ops.object.hide_view_clear()
        bpy.ops.object.select_all(action='DESELECT')
        context.workspace.status_text_set("")

    def invoke(self, context, event):

        self.prep(context)

        context.window_manager.modal_handler_add(self)
        context.workspace.status_text_set("LMB: Add Point          " "Spacebar/Enter: Confirm         "
        "Esc/RMB: Cancel       ""W and S: Offset Surface        " )

        return {"RUNNING_MODAL"}


    def modal(self, context, event):

        if event.type in ['MIDDLEMOUSE' ,'WHEELUPMOUSE', 'WHEELDOWNMOUSE'] or event.alt or event.shift:
            return {'PASS_THROUGH'}

        elif event.type == 'W' and event.value == 'PRESS':
            if self.wrap_mod:
                self.wrap_mod.offset += 0.001

        elif event.type == 'S' and event.value == 'PRESS':
            if self.wrap_mod:
                self.wrap_mod.offset -= 0.001
                if self.wrap_mod.offset < 0:
                    self.wrap_mod.offset = 0

        elif event.type == 'LEFTMOUSE' and event.value == 'RELEASE':
            bpy.ops.mesh.dupli_extrude_cursor('INVOKE_DEFAULT',rotate_source=True)

        elif event.type in {'RET', 'NUMPAD_ENTER', 'SPACE'}:
            self.finish_up(context)
            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE'}:
            self.canceled_op(context)
            return {'CANCELLED'}

        elif event.type in {'ESC'}:
            self.canceled_op(context)
            return {'CANCELLED'}

        return {"RUNNING_MODAL"}



class Handy_Curve_Profile_Preferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    ui_tab_name: bpy.props.StringProperty(default="HCP")
    filepath: bpy.props.StringProperty(name="Mesh File Path",subtype='FILE_PATH',default=str(Path(__file__).parent/"assets"))
    

    def draw(self, context):
        layout = self.layout
        layout.prop(self, 'ui_tab_name',text = 'UI_Tab')
        
        layout.prop(self, 'filepath')

def update_handy_curve_profile_tab_name():
    interface_classes = (HANDY_CURVE_PT_main,)

    for cl in reversed(interface_classes):
        try:
            bpy.utils.unregister_class(cl)
        except:
            pass

    HANDY_CURVE.bl_category = bpy.context.preferences.addons[__package__].preferences.ui_tab_name

    for cl in interface_classes:
        bpy.utils.register_class(cl)



classes = [

    LoadPrevCurveProfileOperator,LoadNextCurveProfileOperator, HandyCurveProfileProperties,HANDY_CURVE_PT_main,ApplyCurveProfileOperator,
    SwitchCurveEndOperator,ApplyBothEndOperator,FinalizeSelectionOperator,Test_Mesh_Before_Append_Operator, Edge_To_Curve_Operator,
    LoadPrevCurveEndOperator,LoadNextCurveEndOperator,ApplyCurveEndOperator,UpdateProxyCurveOperator,ApplyRigidEndOperator,
    ResetCurRadiusOperator,ResetCurTiltOperator,Handy_Curve_Profile_Preferences,Bones_From_Curve_Points_OP,HCP_Append_to_Source_Operator,
    Copy_Curve_Profile_To_Others_Operator,Draw_Curve_On_Surface_OP,
]


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    update_handy_curve_profile_tab_name()

    bpy.types.Scene.handy_curve_profile_props = bpy.props.PointerProperty(
        type=HandyCurveProfileProperties)
    try:
        bpy.app.translations.register(__package__, GetTranslationDict())
    except:
        pass

def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)

    del bpy.types.Scene.handy_curve_profile_props
    try:
        bpy.app.translations.unregister(__package__)
    except:
        pass

if __name__ == "__main__":
    register()
